import { MDXProvider } from "@mdx-js/react";
import { TREE } from "@bolt/config";
import { lazy, Suspense, useEffect, useState } from "react";
import { Route, Routes, useLocation } from "react-router";
import {
  ICompoTree,
  ICustomPage,
  IPropMeta,
  IRoute,
  ITocItem,
  ITree,
} from "../plugin/types";
import ApiTable from "../plugin/common/ApiTable";
import Aside from "../plugin/common/Aside";
import objHash from "object-hash";
import Toc from "../plugin/common/MdToc";
import Nav from "../plugin/common/Nav";

const hash = (str: string) => "" + objHash(str).substring(0, 6);

function App() {
  const location = useLocation();
  const [customPages, setCustomPages] = useState<ICustomPage[]>(
    TREE.config.customPages
  );
  const [compoTree, setCompoTree] = useState<ICompoTree>(TREE.compoTree);
  const [routerList, setRouterList] = useState<IRoute[]>(TREE.routerList);
  const [toc, setToc] = useState<ITocItem[]>([]);
  const [propsTable, setPropsTable] = useState<IPropMeta[]>([]);
  const [isPage, setIsPage] = useState(true);

  /**
   * 客户端刷新通知 vite-server
   */
  useEffect(() => {
    if (import.meta.hot) import.meta.hot.send("client:refresh");
  }, []);

  /**
   * 客户端刷新 vite-server 返回新的 TREE
   */
  useEffect(() => {
    if (import.meta.hot) {
      import.meta.hot.on("client:refresh:response", (response: ITree) => {
        console.log("refresh", response);
        const { routerList, compoTree } = response;
        setRouterList(routerList);
        setCompoTree(compoTree);
      });
    }
  }, []);

  /**
   * vite-server 监听到文件改动
   */
  useEffect(() => {
    if (import.meta.hot) {
      import.meta.hot.on("server:fullUpdate", (response: ITree) => {
        console.log("fullUpdate", response);
        const { routerList, compoTree } = response;
        setRouterList(routerList);
        setCompoTree(compoTree);
      });
    }
  }, []);

  const getCurrent = () => {
    const path = location.pathname.replace(/\//, "");
    const route = routerList.find((i) => i.path === path);
    const compoName = route?.name ?? "";
    setIsPage(compoName === "");
    return { path, route, compoName };
  };

  useEffect(() => {
    const { compoName } = getCurrent();
    if (!!compoName) {
      const { toc, props } = compoTree[compoName];
      setPropsTable(props);
      setToc(toc);
    } else {
      setToc([]);
    }
  }, [compoTree, location.pathname]);

  useEffect(() => {
    setRouterList((p) =>
      p.map((i) => ({ ...i, active: location.pathname.indexOf(i.path) > -1 }))
    );
  }, [compoTree]);

  const H1 = (args: any) => {
    const anchorHash = "anchor-" + hash(args.children);
    const attr = { name: anchorHash };
    return <h1 {...attr}>{args.children}</h1>;
  };

  const H2 = (args: any) => {
    const anchorHash = "anchor-" + hash(args.children);
    const attr = { name: anchorHash };
    return <h2 {...attr}>{args.children}</h2>;
  };

  const lazyLoad = (modulePath: string) => {
    const Module = lazy(() =>
      new Promise((resolve) => setTimeout(resolve, 150)).then(() =>
        import(`./components/${modulePath}/index.mdx`).then((m) => m)
      )
    );
    return (
      <Suspense fallback={<div className="page-loading" />}>
        <MDXProvider components={{ h1: H1, h2: H2 }}>
          <Module />
        </MDXProvider>
      </Suspense>
    );
  };

  const lazyLoadPage = (src: string) => {
    const Module = lazy(() =>
      new Promise((resolve) => setTimeout(resolve, 150)).then(() =>
        import(`./pages/${src}.mdx`).then((m) => m)
      )
    );
    return (
      <Suspense fallback={<div className="page-loading" />}>
        <MDXProvider components={{ h1: H1, h2: H2 }}>
          <Module />
        </MDXProvider>
      </Suspense>
    );
  };

  return (
    <div className="root-container">
      {/* nav */}
      <Nav customPages={customPages} />
      {/* 左侧导航栏 */}
      <Aside
        menus={routerList}
        subTitle={TREE.config.subTitle}
        version={TREE.config.version}
      />
      <main>
        <article className="markdown-container">
          <Routes>
            {/* 自定义页面 */}
            {customPages.map((page: ICustomPage) => (
              <Route
                key={page.path}
                path={page.path}
                element={lazyLoadPage(page.src)}
              />
            ))}
            {/* 组件 */}
            {routerList.map((route: IRoute) => (
              <Route
                key={route.name}
                path={route.path}
                element={lazyLoad(route.name)}
              />
            ))}
          </Routes>
        </article>
        {/* 组件 props 表 */}
        {!isPage && <ApiTable propList={propsTable} />}
        {/* table of content */}
        {toc.length > 0 && <Toc toc={toc} />}
      </main>
    </div>
  );
}

export default App;
